pub enum Reason {
Token(Token),
Name(Token),
Builtin(Token),
MultiplierBug(Token),
VariableReference(Token, &'static str),
}Expand description
This enum records the reason why we think a certain scope has the type it does. It is used for error reporting.
TODO: make a ReasonRef that contains an &Token, and a Borrow impl for it.
This will avoid some cloning.
Variants
Token(Token)
The reason can be explained by pointing at some token
Name(Token)
The scope type was deduced from a named scope’s name; the Token points at that name in
the script.
Builtin(Token)
The scope was supplied by the game engine. The Token points at a key explaining this, for
example the key of an Item or the field key of a trigger or effect in an item.
MultiplierBug(Token)
The vic3 engine evaluates multiplier in add_modifier in root scope, which is probably a
bug. Explain it to the user when it comes up. The Token points at the multiplier key.
VariableReference(Token, &'static str)
The scope type was taken from info about a variable or variable list in the given namespace.
Implementations