enum ScopeEntry {
Backref(usize),
Rootref,
Scope(Scopes, Reason),
Named(usize),
GlobalVar(&'static str, Reason),
GlobalList(&'static str, Reason),
Var(&'static str, Reason),
VarList(&'static str, Reason),
}Expand description
ScopeEntry is a description of what we know of a scope’s type and its connection to other
scopes.
It is used both to look up a scope’s type, and to propagate knowledge about that type backward
to the scope’s source. For example if this is a Rootref, and we find out that this is a
Character, then root must be a Character too.
Variants
Backref(usize)
Backref is for when the current scope is made with prev or this.
It counts as a scope in the chain, for purposes of prev and such, but any updates
to it (such as narrowing of scope types) need to be propagated back to the
real origin of that scope.
The backref number is a relative count into scope_stack, counting backwards from the
entry in question.
It may go past the edge of scope_stack, if the script being analyzed uses prev farther
than we know about
INVARIANT: The usize must not be zero.
Rootref
A Rootref is for when the current scope is made with root. Most of the time,
we also start with this being a Rootref.
Scope(Scopes, Reason)
Reason why we think the Scopes value is what it is.
It’s usually the token that was the cause of the latest narrowing.
Named(usize)
The scope takes its value from a named scope. The usize is an index into the ScopeContext::named vector.
GlobalVar(&'static str, Reason)
The scope takes its value from a global variable
GlobalList(&'static str, Reason)
The scope takes its value from a global variable list
Var(&'static str, Reason)
The scope takes its value from a normal variable
VarList(&'static str, Reason)
The scope takes its value from a variable list
Implementations
Sourceimpl ScopeEntry
impl ScopeEntry
fn deduce<T: Into<Token>>(token: T) -> ScopeEntry