scope_type_from_name

Function scope_type_from_name 

fn scope_type_from_name(name: &str) -> Option<Scopes>
Deduce a scope type from a scope’s name. This leads to better error messages.

It should be limited to names that are so obvious that it’s extremely unlikely that anyone would use them for a different type.